import pickle

from sklearn.compose import ColumnTransformer
from sklearn.datasets import fetch_openml
from sklearn.feature_selection import SelectPercentile, chi2
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler

x, y = fetch_openml("titanic", version=1, as_frame=True, return_X_y=True)

x.columns = [
    "pclass",
    "name",
    "sex",
    "age",
    "sibsp",
    "parch",
    "ticket",
    "fare",
    "cabin",
    "embarked",
    "boat",
    "body",
    "home",
]

print("Dataset Loaded!")

numeric_features = ["age", "fare"]
numeric_transformer = Pipeline(
    steps=[("imputer", SimpleImputer(strategy="median")), ("scaler", StandardScaler())]
)

categorical_features = ["embarked", "sex", "pclass"]
categorical_transformer = Pipeline(
    steps=[
        ("encoder", OneHotEncoder(handle_unknown="ignore")),
        ("selector", SelectPercentile(chi2, percentile=50)),
    ]
)

preprocessor = ColumnTransformer(
    transformers=[
        ("num", numeric_transformer, numeric_features),
        ("cat", categorical_transformer, categorical_features),
    ]
)

clf = LogisticRegression()

pipeline = Pipeline([("feature_preporcessing", preprocessor), ("predictor", clf)])

x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=0)

trans_data = pipeline.fit(x_train, y_train)

# save pipeline
with open("complete_pipeline.pkl", "wb") as pipeline_file:
    pickle.dump(pipeline, pipeline_file)

print("Example model saved!")



import os
import pickle
from contextlib import asynccontextmanager

from fastapi import FastAPI
from loguru import logger

from .api import SimpleAPI, check_model_methods


@asynccontextmanager
async def lifespan(app: FastAPI):
    model_path = os.getenv("MODEL_PATH")

    if model_path is None:
        raise RuntimeError("MODEL_PATH environment variable does not exist.")

    with open(model_path, "rb") as model_file:
        logger.info("Loading deployed model ...")
        app.state.pipeline = pickle.load(model_file)
        try:
            check_model_methods(app.state.pipeline, "predict")
            logger.info("✅ Model loaded")
        except Exception as e:
            message = "The object that was loaded doesn't have `predict` method"
            logger.error(f"{message}: -> {e}")
    yield

    logger.info("👋 Shutting down...")


def serve(simple_api: SimpleAPI):
    """Function that constructs the model API.

    Args:
        simple_api (SimpleAPI): The SimpleAPI object needed for deployment.

    Returns:
        app (FastAPI): The FastAPI application.
    """
    app = FastAPI(lifespan=lifespan)
    api = simple_api
    app.include_router(
        api.routes,
    )
    return app



import uvicorn
from dotenv import load_dotenv
from pydantic import create_model

from sk_serve import SimpleAPI, serve

load_dotenv()

model = create_model(
    "Model",
    pclass=(int, None),
    name=(str, None),
    sex=(str, None),
    age=(float, None),
    sibsp=(int, None),
    parch=(int, None),
    ticket=(str, None),
    fare=(float, None),
    cabin=(str, None),
    embarked=(str, None),
    boat=(int, None),
    body=(float, None),
    home=(str, None),
)

api = SimpleAPI(validation_model=model)

app = serve(api)

if __name__ == "__main__":
    uvicorn.run(
        "example_validation:app",
        host="localhost",
        port=8000,
        log_level="debug",
        reload=True,
    )



import uvicorn
from dotenv import load_dotenv

from sk_serve import SimpleAPI, serve

load_dotenv()

api = SimpleAPI()

app = serve(api)

if __name__ == "__main__":
    uvicorn.run(
        "example:app", host="localhost", port=8000, log_level="debug", reload=True
    )



from typing import Dict, Union

import pandas as pd
from fastapi import APIRouter, Request
from loguru import logger
from pydantic.main import BaseModel


class SimpleAPI:
    """Simple API class that takes pipeline/model path as arguments and defines one inference endpoint for
    simple model deployment. The pipeline must be a Scikit-Learn Pipeline. It can also take a pydantic validation model
    as input in order to validate the input everytime inference is requested.
    """

    def __init__(
        self,
        validation_model: Union[type[BaseModel], None] = None,
    ):
        self.routes = APIRouter()
        self.validation_model = validation_model

        # add our only 2 endpoints
        self.routes.add_api_route("/", getattr(self, "home"), methods=["GET"])
        self.routes.add_api_route(
            "/inference", getattr(self, "inference"), methods=["POST"]
        )

    @staticmethod
    def home() -> Dict[str, str]:
        """Method that returns a message when sending a GET request to the `/` endpoint."""
        home_message = (
            "This is a simple endpoint with a deployed scikit-learn pipeline. \
            Only available endpoints is: [POST] /inference."
        )

        return {"message": home_message}

    async def inference(self, request: Request):
        """Inference method that is used by the inference endpoint. In order to get the prediction
        the deployed pipeline must have the `predict` method.

        Args:
            request (Request): Input data for inference. Currently only one data point at a time is supported.

        Returns:
            dict: The prediction.
        """
        data = await request.json()

        logger.info(data)

        if self.validation_model is not None:
            logger.info("Validation of requerst data ...")
            self.validation_model.model_validate(obj=data)

        x_data = pd.DataFrame(data, index=[0])

        # get predictions
        logger.info("Getting prediction ...")
        preds = request.app.state.pipeline.predict(x_data)

        return {"prediction": preds.item()}


def check_model_methods(model, method: str):
    """Helper function that checks if a class method exits or not.

    Args:
        model: A Scikit-learn model.
        method (str): The name of the respective method.
    """
    try:
        method_name = getattr(model, method)
    except Exception as e:
        logger.error(e)
        raise e

    assert callable(method_name)
